;------------------------------------------------------------------------------------------------  
; The macros in this file give you a simulated line by line debug feature.
; you must define IS_DEBUG to 1 for the macro to work if it is 0 the the macro produces no
; code at all. The macros in this file are:
; Main macros:
; _dbg		defines work areas (place in .data section) 
;                 _dbg also defines the breakpoint lines:
; _dbg 5,25,55  (The debug line i.e. if _asm<mov eax,0> is on editor line 5 that is the line#)
; _fun	     contains the debug function itself (place in a .code section)
; _asm          will pop a msgbox with 3 options:
;               1. yes 		step to next popup
;               2. no 		run to next breakpoint
;               3. cancel 	cancel debug and run to end of job (or abort)
; _asl provides the same function as _asm but creates less code and only shows the line number
;      this can be useful in a "loop" or other piece of code that uses short jumps
;------------------------------------------------------------------------------------------------  
; if you have this instruction: (I use >;; so I can simple do a replace for all >;; and _asm< )
; _asm<mov eax,0>;;
; The popup msgbox's name is the instruction being executed: mov eax,0
; The address and _asm line number are displayed followed by the registers:
; eax, ebx, ecx, edx, ebp, esp, esi, and edi
;
; NOTE:
; A few instructions require special modification notice the extra ! in the following:
; _asm<   .if eax !> 0 and eax !!= 10> ; the leading !s tell the assembler the next char is
; to be taken literally and not to be operated on.  This allows the > and ! to be passed to the
; macro so they can be debugged.
;
; The address in the popup is useful when the windows app crashes and tells you the address 
; where it crashed.  The line number is useful when you have multiple instructions that are
; similar. You can also set breakpoints on a line number.
;
; Other macros:
; _hex          prints a value in hex i.e. _hex<eax> or _hex<mydword> 
; _int          prints a value as an integer _int<eax> or _hex<mydword>
; _trc          puts debug info to a file (caution can be big)
; _tof          turns the trace off and resumes debugging
; _lbl          pops up a name that will notify you when that line is processed:
;               _lbl<entered function xyz>
;------------------------------------------------------------------------------------------------  
; Sample debug code: (note masm32.inc and masm32.lib are required)
; .386
; .model flat, stdcall ; 32 bit memory model
; option casemap :none ; case sensitive
; IS_DEBUG EQU 1
; include \masm32\include\windows.inc
; include \masm32\include\masm32.inc
; ...
; include \your include path\mydebug.inc
; ...
; includelib \masm32\lib\masm32.lib
; ...
; .data
;    len    dw 0
;    inMem  dw 0
;    outmem dw 0
;    inFile db 'myfile.txt',0
; _dbg
; .code
; _fun
; start:
;     _asm<invoke ReadFile>
;     _asm<.if eax !> 0>
;         _asm<invoke mainCode>
;         _asm<invoke WriteFile>
;     _asm<.endif>
;     _asm<mov eax,0>
;     _asm<invoke exitProcess>
;
; Readfile  
; ...
;------------------------------------------End of Help-----------------------------------------  
ifndef wsprintf
wsprintf TEXTEQU <wsprintfA>
endif

_dbg MACRO _brk:VARARG
    LOCAL count 
    IF IS_DEBUG NE 0
        count = 0
        FOR arg, <_brk>
           count = count + 1
           IF count EQ 1
              _BRKDEFS EQU 1
              _mybrks dword arg
           ELSE
                      dword arg
           ENDIF
        ENDM
        ifndef _BRKDEFS
            _mybrks dword 0
        endif
        _mybrkx dword 0
        _myswi  dword 1
        _myretn dword 0
        _myline dword 0
        _myeax  dword 0
        _myebx  dword 0
        _myecx  dword 0
        _myedx  dword 0
        _myebp  dword 0
        _myesp  dword 0
        _myesi  dword 0
        _myedi  dword 0
        _myret  dword 0
        _myptr  dword 0
        _mylen  dword 0
        _myhold dword 0
        _myfile dword 0
        _mybw   dword 0
        _myrnam byte ' eax:',0,' ebx:',0,' ecx:',0,' edx:',0
                byte ' ebp:',0,' esp:',0,' esi:',0,' edi:',0
        _myfnam byte 'mydebug.txt',0
        _myanam byte 'Address ',0
        _mylnam byte ' Line ',0
        _mycrlf byte 13,10,0
        _myins  byte 'Debug',0
        _mywork byte '                      ',0
        _mywrk2 byte '                      ',0
        _mybuff byte 256 dup(0)
        _myhex  byte '%x',0
        _myint  byte '%d',0
        _mystep byte 'Yes=Step, No=Run to breakpoint, Cancel=Go or Abort.',0
        _myques byte 'Debug Interupted.',0
        _myquit byte 'To run without debug click Yes.', 13,10
                byte 'To continue debugging click No.', 13,10
                byte 'To abort (protection) click Cancel.',0
    ENDIF  
ENDM

_trc MACRO
    IF IS_DEBUG NE 0
       mov _myswi,2
    ENDIF  
ENDM

_tof MACRO
    IF IS_DEBUG NE 0
       mov _myswi,1
    ENDIF  
ENDM

_hex MACRO val                 ; reg or other DWORD
    LOCAL _myval, _myvalx
    IF IS_DEBUG NE 0
        .if _myswi == 1 || _myswi == 3
            pushad
            jmp _myvalx
            _myval:
               byte '&val',0
            _myvalx:
            invoke wsprintf, addr _mywork, addr _myhex, &val
            invoke _showdbg,addr _mywork,addr _myval
            popad 
        .endif
    ENDIF  
ENDM

_int MACRO reg                 ; reg or other DWORD
    LOCAL _myreg, _myregx
    IF IS_DEBUG NE 0
        .if _myswi == 1 || _myswi == 3
            pushad
            jmp _myregx
            _myreg:
               byte '&reg',0
            _myregx:
            invoke wsprintf,addr _mywork,addr _myint,&reg
            invoke _showdbg,addr _mywork, addr _myreg
            popad 
        .endif
    ENDIF  
ENDM

_lbl MACRO lbl
    LOCAL _mylbl,_myskp
    IF IS_DEBUG NE 0
        .if _myswi == 1 || _myswi == 3
            jmp _myskp
            _mylbl:
               if @InStr (1, <&instr>, <!'>) EQ 0
                  byte '&lbl',0
               else
                  byte "&lbl",0
               endif
            _myskp:
            pushad
            invoke _showdbg,addr _mylbl,addr _myins
            popad 
        .endif
    ENDIF  
ENDM

_txt MACRO name
    IF IS_DEBUG NE 0
        .if _myswi == 1 || _myswi == 3
            pushad
            invoke _showdbg,&name,addr _myins
            popad 
        .endif
    ENDIF  
ENDM

_asm MACRO instr 
     LOCAL _myskp,_myinst,_mynull,_myinfo
     IF IS_DEBUG NE 0
        .if _myswi == 1 || _myswi == 2 || _myswi == 3
            pushad
            mov _myline,@Line
            call _myskp
            _myinst:
               if @InStr (1, <&instr>, <!'>) EQ 0
                  byte '&instr',0
               else
                  byte "&instr",0
               endif
            _myskp:
            pop _myretn
            lea eax,_myinst
            push eax
            call _mydebug
            popad 
        .endif
     ENDIF  
     &instr
ENDM

_asl MACRO instr 
     LOCAL _myskp,_myinst,_mynull,_myinfo
     IF IS_DEBUG NE 0
        .if _myswi == 1 || _myswi == 2 || _myswi == 3
            pushad
            mov _myline,@Line
            call _myskp
            _myskp:
            pop _myretn
            call _mydebugln
            popad 
        .endif
     ENDIF  
     &instr
ENDM

_fun MACRO
IF IS_DEBUG NE 0
_showdbg proto :DWORD,:DWORD
_mydebugln:    
    pop _myret
    lea eax,_myins
    mov _myptr,eax
    jmp _mydebugsk
_mydebug:    
    pop _myret
    pop _myptr
_mydebugsk:    
    popad
    mov _myeax,eax
    mov _myebx,ebx
    mov _myecx,ecx
    mov _myedx,edx
    mov _myebp,ebp
    mov _myesp,esp
    mov _myesi,esi
    mov _myedi,edi
    pushad
    mov byte ptr _mybuff,0

    invoke szCatStr,addr _mybuff,addr _myanam
    invoke wsprintf,addr _mywork,addr _myhex,_myretn
    invoke szCatStr,addr _mybuff,addr _mywork

    invoke szCatStr,addr _mybuff,addr _mylnam
    invoke wsprintf,addr _mywork,addr _myint,_myline
    invoke szCatStr,addr _mybuff,addr _mywork
    invoke szCatStr,addr _mybuff,addr _mycrlf

    mov ecx,8
    lea ebx,_myrnam
    lea edx,_myeax 
    .repeat
       push ecx
       mov eax,[edx]
       mov _myhold,eax
       push edx
       invoke szCatStr,addr _mybuff,ebx
       invoke wsprintf,addr _mywork,addr _myhex,_myhold
       invoke szCatStr,addr _mybuff,addr _mywork
       pop edx
       add ebx,6
       add edx,4
       pop ecx
    .untilcxz

    invoke szCatStr,addr _mybuff,addr _mycrlf
    invoke szCatStr,addr _mybuff,addr _mycrlf
    invoke szCatStr,addr _mybuff,addr _mystep

    ; we get here for switch 1, 2 and 3 (4 is set if a breakpoint)
    .if _myswi == 3
        lea eax,_mybrks
        .repeat
           mov ebx,[eax]
           .if ebx == 0
              .break
           .endif
           .if ebx == _myline
              mov _myswi,4
              .break
           .endif
           add eax,4
        .until 0
    .endif
    .if _myswi == 1 || _myswi == 4       
       mov _myswi,1
       invoke _showdbg,addr _mybuff, _myptr
    .else
       invoke szCatStr,addr _mybuff,addr _mycrlf
       invoke szCatStr,addr _mybuff,_myptr
       invoke CreateFile,addr _myfnam, ; pointer to name of the file
           GENERIC_WRITE,              ; access (read-write) mode
           NULL,                       ; share mode
           NULL,                       ; pointer to security attributes
           CREATE_ALWAYS,              ; how to create
           FILE_ATTRIBUTE_NORMAL,      ; file attributes
           NULL
       mov _myfile,eax    
       invoke StrLen,addr _mybuff
       mov _mylen,eax
       invoke WriteFile,_myfile,addr _mybuff,_mylen,addr _mybw,NULL
       invoke CloseHandle,_myfile      ; clean up the file
    .endif
    push _myret
    ret
;-----------------------------
_showdbg proc _dbgname:DWORD,_dbgdata:DWORD
       invoke MessageBox, NULL,_dbgname, _dbgdata, MB_YESNOCANCEL
       .if eax == 2
           invoke MessageBox, NULL,addr _myquit, addr _myques, MB_YESNOCANCEL
           .if eax == 6
              mov _myswi,0
           .elseif eax == 7
              mov _myswi,1
           .elseif eax == 2
              mov eax,0
              mov [eax],ebx
           .else
              mov _myswi,1
           .endif
       .elseif eax == 6
           mov _myswi,1
       .elseif eax == 7
           mov _myswi,3
       .else
           mov _myswi,1
       .endif
    ret
_showdbg endp
    ENDIF
ENDM